﻿namespace SnakeWF.Model


//Snake játék modell osztálya.
type SnakeModel() as this =
    class
        //Privát mezők, a let kulcsszóval definiált mezők mindig privátak
        let _random = new System.Random()
        let _timer = new System.Timers.Timer()

        //Táblaméret
        let _tableSize = 50

        //"Változó definiálása"
        let mutable _waiting = 0

        let mutable _isPaused = false
        let mutable _isGameOn = false

        let mutable _gameScore = 0
        let mutable _snakeSize = 0

        let mutable _snake : SnakeWF.Model.Coordinate seq = Seq.empty<SnakeWF.Model.Coordinate>
        let mutable _food : SnakeWF.Model.Coordinate = new SnakeWF.Model.Coordinate(0,0)
        let mutable _snakeDir = SnakeWF.Model.Direction.East


        let TimerTick _ = 
            if (_waiting = 0) then
                this.Move _snakeDir
        
        let IsValid (position : SnakeWF.Model.Coordinate) =
            let isPosValid = position.X >= 0 && position.X < _tableSize && position.Y >= 0 && position.Y < _tableSize
            isPosValid && ( (_snake |> Seq.where(fun x -> x = position) |> Seq.length) < 1)

        //Események definiálása, a DelegateEvent-et példányosítjuk a számunkra szükséges eseményargumentum osztályokkal.
        let updateEvent = new DelegateEvent<System.EventHandler<SnakeWF.Model.UpdateEventArgs> >()
        let gameOverEvent = new DelegateEvent<System.EventHandler<SnakeWF.Model.GameOverEventArgs> >()
        
        //A do parancsban definiált műveletek az objektum létrejöttekor futnak le, kvázi egy konstruktor.
        do
            _timer.Interval <- 150.0
            _timer.Elapsed.Add TimerTick

        //Események publikálása, hogy más nyelven íródott osztályok is feliratkozhassanak rájuk
        [<CLIEvent>]
        member this.Update = updateEvent.Publish
        [<CLIEvent>]
        member this.GameOver = gameOverEvent.Publish


        
        //Publikus tulajdonságok, alapértelmezetten publikusak, azonban a private kulcsszóval priváttá tehetőek
        member this.TableSize = _tableSize
        member this.isGameOn = _isGameOn

        //Publikus metódusok

        //Kígyó mozgása
        member this.Move (direction : SnakeWF.Model.Direction) =
            
            if (int direction + int _snakeDir) = 0 || _waiting > 0 then 
                () //Unit értékkel térünk vissza, ez azt jelenti, hogy nincs visszaadott érték
            else
                _waiting <- _waiting + 1 
                let tmp = _snake |> Seq.last
                let head = new Coordinate(tmp.X,tmp.Y)

                //Az irányt mintaillesztéssel kérhetjük le
                match direction with
                    | SnakeWF.Model.Direction.North -> head.Y <- head.Y - 1
                    | SnakeWF.Model.Direction.East -> head.X <- head.X + 1
                    | SnakeWF.Model.Direction.South -> head.Y <- head.Y + 1
                    | SnakeWF.Model.Direction.West -> head.X <- head.X - 1
                    | _ -> ()

                if not( IsValid head ) then
                    _waiting <- _waiting - 1 
                    this.OnGameOver
                else
                    if (head = _food) then
                        _food.X <- _random.Next(_tableSize)
                        _food.Y <- _random.Next(_tableSize)
                        _gameScore <- _gameScore + 1
                    else
                        _snake <- _snake |> Seq.skip 1

                    _snake <- Seq.append _snake [head]
                    _snakeDir <- direction
                    //updateEvent kiváltása, a paramétereket egy tömbben adjuk meg, először a küldő, aztán az esemény-
                    //argumentumok, a box kulcsszóval "dobozolhatjuk" be a paramétereket, ez nem csak boxingot végez,
                    //hanem Object típusúvá kasztolja az objektumokat.
                    updateEvent.Trigger( [| box this ; box (new UpdateEventArgs(_food, (Seq.toList _snake))) |] )
                    _waiting <- _waiting - 1

        //Új játék indítása
        member this.NewGame() = 
            _gameScore <- 0
            _waiting <- 0
            _isGameOn <- true

            _food <- SnakeWF.Model.Coordinate(_random.Next(_tableSize), _random.Next(_tableSize))
            //A Seq.init metódussal és anonim függvények segítségével könnyedén inicializálhatjuk a kígyónk pontjait,
            //első paraméterként megadjuk, milyen hosszú lesz, a második paraméterként pedig az anonim függvénnyel
            //megmondhatjuk, hogy ezek a koordináták hol helyezkednek el a táblán.
            _snake <- Seq.init<SnakeWF.Model.Coordinate> 5 (fun i -> (new SnakeWF.Model.Coordinate(_tableSize / 2 + i,_tableSize / 2)))
            _snakeSize <- Seq.length _snake
            _snakeDir <- SnakeWF.Model.Direction.East
            updateEvent.Trigger([| box this ; box (new UpdateEventArgs(_food, (Seq.toList _snake))) |])
            _timer.Start()            

        
        member private this.OnGameOver =
            _timer.Stop()
            gameOverEvent.Trigger([| box this; box (new GameOverEventArgs(_gameScore)) |]) //Esemény kiváltása

        member this.PauseOrResume() =
            if(_isPaused) then
                _timer.Start()
                _isPaused <- false
            else
                _timer.Stop()
                _isPaused <- true

    end